self.__BUILD_MANIFEST=function(s,c,a,e,t,f,b,u,n,o,j,r,i,k,h){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":["static/chunks/pages/index-444f1804401f04ea.js"],"/_error":["static/chunks/pages/_error-c66a4e8afc46f17b.js"],"/clusters":["static/chunks/pages/clusters-9e0df5442b04f7d0.js"],"/clusters/[cluster]":[s,c,a,f,b,j,o,e,t,u,r,n,i,k,h,"static/chunks/6856-5c94d394259cdb6e.js","static/chunks/1871-81adbc5f25dff78f.js","static/chunks/pages/clusters/[cluster]-1704039ccaf997cf.js"],"/clusters/[cluster]/[job]":[s,c,a,f,e,t,n,"static/chunks/pages/clusters/[cluster]/[job]-ac4a217f17b087cb.js"],"/config":["static/chunks/pages/config-dfb9bf07b13045f4.js"],"/infra":["static/chunks/pages/infra-c84a3b8a9d599b02.js"],"/infra/[context]":["static/chunks/pages/infra/[context]-9b6e47c2e8b485a2.js"],"/jobs":["static/chunks/pages/jobs-7eee823559e5cf9f.js"],"/jobs/pools/[pool]":[s,c,a,b,o,e,t,u,"static/chunks/pages/jobs/pools/[pool]-e020fd69dbe76cea.js"],"/jobs/[job]":[s,c,a,f,b,o,e,t,u,n,"static/chunks/pages/jobs/[job]-eb5822dac0c9509b.js"],"/users":["static/chunks/pages/users-2b172f13f8538a7a.js"],"/volumes":["static/chunks/pages/volumes-d2af9d22e87cc4ba.js"],"/workspace/new":["static/chunks/pages/workspace/new-3f88a1c7e86a3f86.js"],"/workspaces":["static/chunks/pages/workspaces-1891376c08050940.js"],"/workspaces/[name]":[s,c,a,f,b,j,e,t,u,r,n,i,k,h,"static/chunks/1141-c3c10e2c6ed71a8f.js","static/chunks/pages/workspaces/[name]-bbfe5860c93470fd.js"],sortedPages:["/","/_app","/_error","/clusters","/clusters/[cluster]","/clusters/[cluster]/[job]","/config","/infra","/infra/[context]","/jobs","/jobs/pools/[pool]","/jobs/[job]","/users","/volumes","/workspace/new","/workspaces","/workspaces/[name]"]}}("static/chunks/616-3d59f75e2ccf9321.js","static/chunks/6130-2be46d70a38f1e82.js","static/chunks/5739-d67458fcb1386c92.js","static/chunks/6989-01359c57e018caa4.js","static/chunks/3850-ff4a9a69d978632b.js","static/chunks/7411-b15471acd2cba716.js","static/chunks/1272-1ef0bf0237faccdb.js","static/chunks/8969-4ed9236db997b42b.js","static/chunks/9353-cff34f7e773b2e2b.js","static/chunks/6212-7bd06f60ba693125.js","static/chunks/7359-c8d04e06886000b3.js","static/chunks/6990-f6818c84ed8f1c86.js","static/chunks/4282-49b2065b7336e496.js","static/chunks/6601-06114c982db410b6.js","static/chunks/7615-80aa7b09f45a86d2.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();